local addonName, addon = ...

local grid, settings = CreateFrame('Frame', nil, UIParent)						-- Must use UIParent for uiscale
addon.grid = grid

--[[----------------------------------------------------------------------------
Resolution and windowed mode support
------------------------------------------------------------------------------]]
local HALF_SCREEN_HEIGHT, HALF_SCREEN_WIDTH = 384, 512

do
	local function OnUpdate(self)
		self:SetScript('OnUpdate', nil)

		local width, height = self:GetSize()

		HALF_SCREEN_HEIGHT, HALF_SCREEN_WIDTH = height * 0.5, width * 0.5

		if settings then
			addon.UpdateCenterLines()
			addon.DrawGrid()
		end
	end

	grid:SetScript('OnSizeChanged', function(self)
		self:SetScript('OnUpdate', OnUpdate)
	end)

	grid:SetScript('OnUpdate', OnUpdate)
end

--[[----------------------------------------------------------------------------
Line support
------------------------------------------------------------------------------]]
local drawnLines, erasedLines = { }, { }

grid:SetFrameStrata('BACKGROUND')
grid:SetFrameLevel(0)
grid:SetAllPoints(nil)
grid:Hide()

local backdrop = WorldFrame:CreateTexture(nil, 'BORDER', nil, -8)				-- Not visible with 'BACKGROUND'
backdrop:SetTexture([[Interface\BUTTONS\WHITE8X8]])
backdrop:SetBlendMode('DISABLE')
backdrop:SetAllPoints(nil)
backdrop:Hide()

local function CreateLine(drawLayer, subLayer)
	local line = grid:CreateLine(nil, drawLayer, nil, subLayer)
	line:SetBlendMode('DISABLE')
	line:SetSnapToPixelGrid(false)
	line:SetTexelSnappingBias(0)
	return line
end

local centerLineX = CreateLine('BACKGROUND', -7)
centerLineX:SetStartPoint('TOP', grid, 0, 0)
centerLineX:SetEndPoint('BOTTOM', grid, 0, 0)

local centerLineY = CreateLine('BACKGROUND', -7)
centerLineY:SetStartPoint('LEFT', grid, 0, 0)
centerLineY:SetEndPoint('RIGHT', grid, 0, 0)

local function DrawLine(relPoint1, relPoint2, offsetX, offsetY, thickness)
	local line = erasedLines[#erasedLines]
	if line then
		erasedLines[#erasedLines] = nil
	else
		if settings.overlay then
			line = CreateLine('OVERLAY', 6)
		else
			line = CreateLine('BACKGROUND', -8)
		end
		local color = settings.gridColor
		line:SetColorTexture(color[1], color[2], color[3], color[4])
	end
	drawnLines[#drawnLines + 1] = line
	line:SetStartPoint(relPoint1, grid, offsetX, offsetY)
	line:SetEndPoint(relPoint2, grid, offsetX, offsetY)
	line:SetThickness(thickness)
	line:Show()
end

local function DrawHorzLines(start, step)
	local thickness = settings.gridThickness
	for offset = start, HALF_SCREEN_HEIGHT, step do
		DrawLine('LEFT', 'RIGHT', 0, offset, thickness)
		DrawLine('LEFT', 'RIGHT', 0, -offset, thickness)
	end
end

local function DrawVertLines(start, step)
	local thickness = settings.gridThickness
	for offset = start, HALF_SCREEN_WIDTH, step do
		DrawLine('TOP', 'BOTTOM', offset, 0, thickness)
		DrawLine('TOP', 'BOTTOM', -offset, 0, thickness)
	end
end

local function ForGrid(func, ...)
	for index = 1, #drawnLines do
		func(drawnLines[index], ...)
	end
	for index = 1, #erasedLines do
		func(erasedLines[index], ...)
	end
end

--[[----------------------------------------------------------------------------
Global to addon
------------------------------------------------------------------------------]]
function addon.ApplyColor(field)
	local color = settings[field or "gridColor"]
	if field == "backdropColor" then
		backdrop:SetVertexColor(color[1], color[2], color[3], color[4])
	elseif field == "centerColor" then
		centerLineX:SetColorTexture(color[1], color[2], color[3], color[4])
		centerLineY:SetColorTexture(color[1], color[2], color[3], color[4])
	else
		ForGrid(backdrop.SetColorTexture, color[1], color[2], color[3], color[4])
		if not field then
			addon.ApplyColor("backdropColor")
			addon.ApplyColor("centerColor")
		end
	end
end

function addon.DrawGrid()
	local numErased = #erasedLines
	for index = #drawnLines, 1, -1 do
		numErased = numErased + 1
		erasedLines[numErased] = drawnLines[index]
		drawnLines[index]:Hide()
		drawnLines[index] = nil
	end
	local spacingX, spacingY = settings.spacingX, settings[settings.independent and "spacingY" or "spacingX"]
	if settings.absolute then
		DrawVertLines(spacingX, spacingX)
		DrawHorzLines(spacingY, spacingY)
	else
		DrawVertLines(spacingX + settings.centerThickness - 1, spacingX + settings.gridThickness - 1)
		DrawHorzLines(spacingY + settings.centerThickness - 1, spacingY + settings.gridThickness - 1)
	end
end

function addon.UpdateCenterLines()
	centerLineX:SetThickness(settings.centerThickness)
	centerLineY:SetThickness(settings.centerThickness)
end

do
	local overlay, window = false, 1

	function addon.UpdateState()
		backdrop:SetShown((settings.backdropIndependent or not settings.hide) and settings.backdropEnable)
		grid:SetShown(not settings.hide)
		if settings.window ~= window then
			window = settings.window
			local frame = window == 2 and UIParent or window == 3 and WorldFrame or nil
			backdrop:ClearAllPoints()
			backdrop:SetAllPoints(frame)
			grid:ClearAllPoints()
			grid:SetAllPoints(frame)
		end
		if settings.overlay ~= overlay then
			overlay = settings.overlay
			local layer, level, strata, subLayer
			if overlay then
				layer, level, strata, subLayer = 'OVERLAY', 127, 'TOOLTIP', 7
			else
				layer, level, strata, subLayer = 'BACKGROUND', 0, 'BACKGROUND', -7
			end
			grid:SetFrameStrata(strata)
			grid:SetFrameLevel(level)
			ForGrid(backdrop.SetDrawLayer, layer, subLayer - 1)
			centerLineX:SetDrawLayer(layer, subLayer)
			centerLineY:SetDrawLayer(layer, subLayer)
		end
	end
end

--[[----------------------------------------------------------------------------
Initialize
------------------------------------------------------------------------------]]
local function UpdateProfile()
	settings = addon.db.profile
	addon.ApplyColor()
	addon.UpdateState()
	addon.UpdateCenterLines()
	addon.DrawGrid()
end

grid:SetScript('OnEvent', function(self, event, name)
	if addonName ~= name then return end
	self:UnregisterEvent(event)

	self:SetScript('OnEvent', self:GetScript('OnSizeChanged'))
	self:RegisterEvent('DISPLAY_SIZE_CHANGED')

	addon.db = LibStub('AceDB-3.0'):New(addonName .. "Settings", { profile = addon.defaultSettings }, 'Default')
	UpdateProfile()
	CONFIGMODE_CALLBACKS[addonName] = addon.ConfigMode
	addon.db.RegisterCallback(addon, 'OnProfileChanged', UpdateProfile)
	addon.db.RegisterCallback(addon, 'OnProfileCopied', UpdateProfile)
	addon.db.RegisterCallback(addon, 'OnProfileReset', UpdateProfile)
end)
grid:RegisterEvent('ADDON_LOADED')
